@@ -29,16 +29,12 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include <sys/cdefs.h>
-#if defined(LIBC_SCCS) && !defined(lint)
-__RCSID("$NetBSD: pidfile.c,v 1.14 2016/04/12 20:40:43 roy Exp $");
-#endif
-
 #include <sys/param.h>
 
 #include <errno.h>
 #include <fcntl.h>
 #include <inttypes.h>
+#include <limits.h>
 #include <paths.h>
 #include <stdbool.h>
 #include <stdlib.h>
@@ -47,8 +43,8 @@ __RCSID("$NetBSD: pidfile.c,v 1.14 2016/04/12 20:40:43 roy Exp $");
 #include <unistd.h>
 
 #include <sys/file.h>	/* for flock(2) */
-#include "../config.h"
-#include "../defs.h"
+#include "config.h"
+#include "defs.h"
 
 static pid_t pidfile_pid;
 static char pidfile_path[PATH_MAX];
@@ -209,10 +205,10 @@ pidfile_lock(const char *path)
 		opts |=	O_EXLOCK;
 #endif
 		if ((fd = open(path, opts, 0644)) == -1)
-			return -1;
+			goto return_pid;
 #ifndef O_CLOEXEC
 		if ((opts = fcntl(fd, F_GETFD)) == -1 ||
-		    fctnl(fd, F_SETFL, opts | FD_CLOEXEC) == -1)
+		    fcntl(fd, F_SETFL, opts | FD_CLOEXEC) == -1)
 		{
 			int error = errno;
 
@@ -233,6 +229,8 @@ pidfile_lock(const char *path)
 			fd = -1;
 		}
 #endif
+
+return_pid:
 		if (fd == -1) {
 			pid_t pid;
 
