@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -151,8 +151,10 @@
 #define IRT_DEFAULT		86400
 #define IRT_MINIMUM		600
 
-#define DHCP6_RAND_MIN		-100
-#define DHCP6_RAND_MAX		100
+/* These should give -.1 to .1 randomness */
+#define	DHCP6_RAND_MIN		-100
+#define	DHCP6_RAND_MAX		100
+#define	DHCP6_RAND_DIV		1000.0f
 
 enum DH6S {
 	DH6S_INIT,
@@ -167,8 +169,6 @@ enum DH6S {
 	DH6S_RENEW_REQUESTED,
 	DH6S_PROBE,
 	DH6S_DELEGATED,
-	DH6S_TIMEDOUT,
-	DH6S_ITIMEDOUT,
 	DH6S_RELEASE,
 	DH6S_RELEASED
 };
@@ -177,16 +177,18 @@ struct dhcp6_state {
 	enum DH6S state;
 	struct timespec started;
 
-	/* Message retransmission timings */
-	struct timespec RT;
+	/* Message retransmission timings in seconds */
 	unsigned int IMD;
 	unsigned int RTC;
-	time_t IRT;
+	unsigned int IRT;
 	unsigned int MRC;
-	time_t MRT;
+	unsigned int MRT;
 	void (*MRCcallback)(void *);
-	time_t sol_max_rt;
-	time_t inf_max_rt;
+	unsigned int sol_max_rt;
+	unsigned int inf_max_rt;
+	unsigned int RT;	/* retransmission timer in milliseconds
+				 * maximal RT is 1 day + RAND,
+				 * so should be enough */
 
 	struct dhcp6_message *send;
 	size_t send_len;
@@ -208,6 +210,7 @@ struct dhcp6_state {
 	char leasefile[sizeof(LEASEFILE6) + IF_NAMESIZE + (IF_SSIDLEN * 4) +3];
 	const char *reason;
 	uint16_t lerror; /* Last error received from DHCPv6 reply. */
+	bool has_no_binding;
 	struct authstate auth;
 };
 
@@ -219,6 +222,8 @@ struct dhcp6_state {
 	(D6_CSTATE((ifp)) &&						       \
 	D6_CSTATE((ifp))->reason && dhcp6_dadcompleted((ifp)))
 
+int dhcp6_openudp(unsigned int, struct in6_addr *);
+void dhcp6_recvmsg(struct dhcpcd_ctx *, struct msghdr *, struct ipv6_addr *);
 void dhcp6_printoptions(const struct dhcpcd_ctx *,
     const struct dhcp_opt *, size_t);
 const struct ipv6_addr *dhcp6_iffindaddr(const struct interface *ifp,
@@ -236,7 +241,6 @@ void dhcp6_handleifa(int, struct ipv6_addr *, pid_t);
 int dhcp6_dadcompleted(const struct interface *);
 void dhcp6_abort(struct interface *);
 void dhcp6_drop(struct interface *, const char *);
-int dhcp6_dump(struct interface *);
 #endif /* DHCP6 */
 
 #endif /* DHCP6_H */
