@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * Linux interface driver for dhcpcd
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -81,6 +81,7 @@
 #include "ipv6.h"
 #include "ipv6nd.h"
 #include "logerr.h"
+#include "privsep.h"
 #include "route.h"
 #include "sa.h"
 
@@ -108,6 +109,24 @@ int if_getssid_wext(const char *ifname, uint8_t *ssid);
 #define	SOL_NETLINK	270
 #endif
 
+/* Someone should fix kernel headers for clang alignment warnings . */
+#undef IFA_RTA
+#define IFA_RTA(r)  ((struct rtattr *)(void *)(((char *)(r)) \
+	+ NLMSG_ALIGN(sizeof(struct ifaddrmsg))))
+#undef IFLA_RTA
+#define IFLA_RTA(r)  ((struct rtattr *)(void *)(((char *)(r)) \
+	+ NLMSG_ALIGN(sizeof(struct ifinfomsg))))
+#undef NLMSG_NEXT
+#define NLMSG_NEXT(nlh, len)	 ((len) -= NLMSG_ALIGN((nlh)->nlmsg_len), \
+	(struct nlmsghdr *)(void *)(((char *)(nlh)) \
+	+ NLMSG_ALIGN((nlh)->nlmsg_len)))
+#undef RTM_RTA
+#define RTM_RTA(r) (void *)(((char *)(r)) + NLMSG_ALIGN(sizeof(struct rtmsg)))
+#undef RTA_NEXT
+#define RTA_NEXT(rta, attrlen)	((attrlen) -= RTA_ALIGN((rta)->rta_len), \
+	(struct rtattr *)(void *)(((char *)(rta)) \
+	+ RTA_ALIGN((rta)->rta_len)))
+
 struct priv {
 	int route_fd;
 	uint32_t route_pid;
@@ -229,11 +248,18 @@ check_proc_hex(const char *path, unsigned int *value)
 }
 
 static ssize_t
-if_writepathuint(const char *path, unsigned int val)
+if_writepathuint(struct dhcpcd_ctx *ctx, const char *path, unsigned int val)
 {
 	FILE *fp;
 	ssize_t r;
 
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP)
+		return ps_root_writepathuint(ctx, path, val);
+#else
+	UNUSED(ctx);
+#endif
+
 	fp = fopen(path, "w");
 	if (fp == NULL)
 		return -1;
@@ -261,7 +287,7 @@ if_init(struct interface *ifp)
 		return errno == ENOENT ? 0 : -1;
 	if (n == 1)
 		return 0;
-	return if_writepathuint(path, 1) == -1 ? -1 : 0;
+	return if_writepathuint(ifp->ctx, path, 1) == -1 ? -1 : 0;
 }
 
 int
@@ -334,7 +360,7 @@ if_vlanid(const struct interface *ifp)
 	return (unsigned short)v.u.VID;
 }
 
-static int
+int
 if_linksocket(struct sockaddr_nl *nl, int protocol)
 {
 	int fd;
@@ -407,6 +433,23 @@ if_closesockets_os(struct dhcpcd_ctx *ctx)
 	}
 }
 
+int
+if_setmac(struct interface *ifp, void *mac, uint8_t maclen)
+{
+	struct ifreq ifr = {
+		.ifr_hwaddr.sa_family = ifp->family,
+	};
+
+	if (ifp->hwlen != maclen || maclen > sizeof(ifr.ifr_hwaddr.sa_data)) {
+		errno = EINVAL;
+		return -1;
+	}
+
+	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
+	memcpy(ifr.ifr_hwaddr.sa_data, mac, maclen);
+	return if_ioctl(ifp->ctx, SIOCSIFHWADDR, &ifr, sizeof(ifr));
+}
+
 int
 if_carrier(struct interface *ifp)
 {
@@ -416,7 +459,7 @@ if_carrier(struct interface *ifp)
 	return ifp->flags & IFF_RUNNING ? LINK_UP : LINK_DOWN;
 }
 
-static int
+int
 if_getnetlink(struct dhcpcd_ctx *ctx, struct iovec *iov, int fd, int flags,
     int (*cb)(struct dhcpcd_ctx *, void *, struct nlmsghdr *), void *cbarg)
 {
@@ -504,7 +547,7 @@ if_copyrt(struct dhcpcd_ctx *ctx, struct rt *rt, struct nlmsghdr *nlm)
 	if (rtm->rtm_type == RTN_UNREACHABLE)
 		rt->rt_flags |= RTF_REJECT;
 
-	rta = (struct rtattr *)RTM_RTA(rtm);
+	rta = RTM_RTA(rtm);
 	len = RTM_PAYLOAD(nlm);
 	for (; RTA_OK(rta, len); rta = RTA_NEXT(rta, len)) {
 		sa = NULL;
@@ -610,6 +653,11 @@ link_route(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 	}
 
 	/* Ignore messages we sent. */
+#ifdef PRIVSEP
+	if (ctx->ps_root_pid != 0 &&
+	    nlm->nlmsg_pid == (uint32_t)ctx->ps_root_pid)
+		return 0;
+#endif
 	priv = (struct priv *)ctx->priv;
 	if (nlm->nlmsg_pid == priv->route_pid)
 		return 0;
@@ -644,10 +692,18 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		return -1;
 	}
 
-	/* Ignore messages we sent. */
-	priv = (struct priv*)ctx->priv;
-	if (nlm->nlmsg_pid == priv->route_pid)
-		return 0;
+	/* Ignore address deletions from ourself.
+	 * We need to process address flag changes though. */
+	if (nlm->nlmsg_type == RTM_DELADDR) {
+#ifdef PRIVSEP
+		if (ctx->ps_root_pid != 0 &&
+		    nlm->nlmsg_pid == (uint32_t)ctx->ps_root_pid)
+			return 0;
+#endif
+		priv = (struct priv*)ctx->priv;
+		if (nlm->nlmsg_pid == priv->route_pid)
+			return 0;
+	}
 
 	ifa = NLMSG_DATA(nlm);
 	if ((ifp = if_findindex(ctx->ifaces, ifa->ifa_index)) == NULL) {
@@ -655,7 +711,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		 * so it's not really an error */
 		return 1;
 	}
-	rta = (struct rtattr *)IFA_RTA(ifa);
+	rta = IFA_RTA(ifa);
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifa));
 	switch (ifa->ifa_family) {
 #ifdef INET
@@ -761,7 +817,7 @@ link_neigh(struct dhcpcd_ctx *ctx, __unused struct interface *ifp,
 		return -1;
 
 	r = NLMSG_DATA(nlm);
-	rta = (struct rtattr *)RTM_RTA(r);
+	rta = RTM_RTA(r);
 	len = RTM_PAYLOAD(nlm);
         if (r->ndm_family == AF_INET6) {
 		bool reachable;
@@ -903,6 +959,27 @@ if_handlelink(struct dhcpcd_ctx *ctx)
 	    &link_netlink, NULL);
 }
 
+#ifdef PRIVSEP
+static bool
+if_netlinkpriv(int protocol, struct nlmsghdr *nlm)
+{
+
+	if (protocol != NETLINK_ROUTE)
+		return false;
+
+	switch(nlm->nlmsg_type) {
+	case RTM_NEWADDR:	/* FALLTHROUGH */
+	case RTM_DELADDR:	/* FALLTHROUGH */
+	case RTM_NEWROUTE:	/* FALLTHROUGH */
+	case RTM_DELROUTE:	/* FALLTHROUGH */
+	case RTM_NEWLINK:
+		return true;
+	default:
+		return false;
+	}
+}
+#endif
+
 static int
 if_sendnetlink(struct dhcpcd_ctx *ctx, int protocol, struct nlmsghdr *hdr,
     int (*cb)(struct dhcpcd_ctx *, void *, struct nlmsghdr *), void *cbarg)
@@ -914,23 +991,37 @@ if_sendnetlink(struct dhcpcd_ctx *ctx, int protocol, struct nlmsghdr *hdr,
 	    .msg_name = &snl, .msg_namelen = sizeof(snl),
 	    .msg_iov = &iov, .msg_iovlen = 1
 	};
+	bool use_rfd;
 
-	if ((s = if_linksocket(&snl, protocol)) == -1)
-		return -1;
-
-#ifdef NETLINK_GET_STRICT_CHK
-	if (hdr->nlmsg_type == RTM_GETADDR) {
-		int on = 1;
-
-		if (setsockopt(s, SOL_NETLINK, NETLINK_GET_STRICT_CHK,
-		    &on, sizeof(on)) == -1 && errno != ENOPROTOOPT)
-			logerr("%s: NETLINK_GET_STRICT_CHK", __func__);
-	}
-#endif
+	use_rfd = (protocol == NETLINK_ROUTE && hdr->nlmsg_type != RTM_GETADDR);
 
 	/* Request a reply */
 	hdr->nlmsg_flags |= NLM_F_ACK;
 	hdr->nlmsg_seq = (uint32_t)++ctx->seq;
+
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP && if_netlinkpriv(protocol, hdr))
+		return (int)ps_root_sendnetlink(ctx, protocol, &msg);
+#endif
+
+	if (use_rfd) {
+		struct priv *priv = (struct priv *)ctx->priv;
+
+		s = priv->route_fd;
+	} else {
+		if ((s = if_linksocket(&snl, protocol)) == -1)
+			return -1;
+#ifdef NETLINK_GET_STRICT_CHK
+		if (hdr->nlmsg_type == RTM_GETADDR) {
+			int on = 1;
+
+			if (setsockopt(s, SOL_NETLINK, NETLINK_GET_STRICT_CHK,
+			    &on, sizeof(on)) == -1 && errno != ENOPROTOOPT)
+				logerr("%s: NETLINK_GET_STRICT_CHK", __func__);
+		}
+#endif
+	}
+
 	if (sendmsg(s, &msg, 0) != -1) {
 		unsigned char buf[16 * 1024];
 		struct iovec riov = {
@@ -941,7 +1032,9 @@ if_sendnetlink(struct dhcpcd_ctx *ctx, int protocol, struct nlmsghdr *hdr,
 		r = if_getnetlink(ctx, &riov, s, 0, cb, cbarg);
 	} else
 		r = -1;
-	close(s);
+
+	if (!use_rfd)
+		close(s);
 	return r;
 }
 
@@ -1284,7 +1377,7 @@ _if_addressexists(__unused struct dhcpcd_ctx *ctx,
 	ifa = NLMSG_DATA(nlm);
 	if (ifa->ifa_index != ia->ifa_ifindex || ifa->ifa_family != AF_INET)
 		return 0;
-	rta = (struct rtattr *)IFA_RTA(ifa);
+	rta = IFA_RTA(ifa);
 	len = NLMSG_PAYLOAD(nlm, sizeof(*ifa));
 	for (; RTA_OK(rta, len); rta = RTA_NEXT(rta, len)) {
 		switch (rta->rta_type) {
@@ -1316,7 +1409,6 @@ if_addressexists(struct interface *ifp, struct in_addr *addr)
 }
 #endif
 
-
 struct nlmr
 {
 	struct nlmsghdr hdr;
@@ -1710,8 +1802,7 @@ if_address6(unsigned char cmd, const struct ipv6_addr *ia)
 #endif
 		}
 #elif IFA_F_MANAGETEMPADDR
-		if (ia->flags & IPV6_AF_AUTOCONF &&
-		    ipv6_ifidlen(ia->iface) + ia->prefix_len == 128)
+		if (ia->flags & IPV6_AF_AUTOCONF && IA6_CANAUTOCONF(ia))
 			flags |= IFA_F_MANAGETEMPADDR;
 #endif
 #ifdef IFA_F_NOPREFIXROUTE
@@ -1866,7 +1957,7 @@ if_setup_inet6(const struct interface *ifp)
 	snprintf(path, sizeof(path), "%s/%s/autoconf", p_conf, ifp->name);
 	ra = check_proc_int(path);
 	if (ra != 1 && ra != -1) {
-		if (if_writepathuint(path, 0) == -1)
+		if (if_writepathuint(ifp->ctx, path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 
@@ -1878,7 +1969,7 @@ if_setup_inet6(const struct interface *ifp)
 		if (errno != ENOENT)
 			logerr("%s: %s", __func__, path);
 	} else if (ra != 0) {
-		if (if_writepathuint(path, 0) == -1)
+		if (if_writepathuint(ifp->ctx, path, 0) == -1)
 			logerr("%s: %s", __func__, path);
 	}
 }
@@ -1888,19 +1979,20 @@ if_applyra(const struct ra *rap)
 {
 	char path[256];
 	const char *ifname = rap->iface->name;
+	struct dhcpcd_ctx *ctx = rap->iface->ctx;
 	int error = 0;
 
 	snprintf(path, sizeof(path), "%s/%s/hop_limit", p_conf, ifname);
-	if (if_writepathuint(path, rap->hoplimit) == -1)
+	if (if_writepathuint(ctx, path, rap->hoplimit) == -1)
 		error = -1;
 
 	snprintf(path, sizeof(path), "%s/%s/retrans_time_ms", p_neigh, ifname);
-	if (if_writepathuint(path, rap->retrans) == -1)
+	if (if_writepathuint(ctx, path, rap->retrans) == -1)
 		error = -1;
 
 	snprintf(path, sizeof(path), "%s/%s/base_reachable_time_ms",
 	    p_neigh, ifname);
-	if (if_writepathuint(path, rap->reachable) == -1)
+	if (if_writepathuint(ctx, path, rap->reachable) == -1)
 		error = -1;
 
 	return error;
