@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -66,7 +66,7 @@
 #define O_NOIPV6RS		O_BASE + 5
 #define O_IPV6RA_FORK		O_BASE + 6
 #define O_LINK_RCVBUF		O_BASE + 7
-// unused			O_BASE + 8
+#define O_ANONYMOUS		O_BASE + 8
 #define O_NOALIAS		O_BASE + 9
 #define O_IA_NA			O_BASE + 10
 #define O_IA_TA			O_BASE + 11
@@ -106,6 +106,7 @@
 #define O_LASTLEASE_EXTEND	O_BASE + 46
 #define O_INACTIVE		O_BASE + 47
 #define	O_MUDURL		O_BASE + 48
+#define	O_MSUSERCLASS		O_BASE + 49
 
 const struct option cf_options[] = {
 	{"background",      no_argument,       NULL, 'b'},
@@ -129,6 +130,9 @@ const struct option cf_options[] = {
 	{"inform6",         optional_argument, NULL, O_INFORM6},
 	{"timeout",         required_argument, NULL, 't'},
 	{"userclass",       required_argument, NULL, 'u'},
+#ifndef SMALL
+	{"msuserclass",     required_argument, NULL, O_MSUSERCLASS},
+#endif
 	{"vendor",          required_argument, NULL, 'v'},
 	{"waitip",          optional_argument, NULL, 'w'},
 	{"exit",            no_argument,       NULL, 'x'},
@@ -161,6 +165,7 @@ const struct option cf_options[] = {
 	{"oneshot",         no_argument,       NULL, '1'},
 	{"ipv4only",        no_argument,       NULL, '4'},
 	{"ipv6only",        no_argument,       NULL, '6'},
+	{"anonymous",       no_argument,       NULL, O_ANONYMOUS},
 	{"arping",          required_argument, NULL, O_ARPING},
 	{"destination",     required_argument, NULL, O_DESTINATION},
 	{"fallback",        required_argument, NULL, O_FALLBACK},
@@ -286,6 +291,7 @@ add_environ(char ***array, const char *value, int uniq)
 #define PARSE_STRING_NULL	1
 #define PARSE_HWADDR		2
 #define parse_string(a, b, c) parse_str((a), (b), (c), PARSE_STRING)
+#define parse_nstring(a, b, c) parse_str((a), (b), (c), PARSE_STRING_NULL)
 #define parse_hwaddr(a, b, c) parse_str((a), (b), (c), PARSE_HWADDR)
 static ssize_t
 parse_str(char *sbuf, size_t slen, const char *str, int flags)
@@ -722,7 +728,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		ARG_REQUIRED;
 		if (ifo->script != default_script)
 			free(ifo->script);
-		s = parse_str(NULL, 0, arg, PARSE_STRING_NULL);
+		s = parse_nstring(NULL, 0, arg);
 		if (s == 0) {
 			ifo->script = NULL;
 			break;
@@ -733,7 +739,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			logerr(__func__);
 			return -1;
 		}
-		s = parse_str(ifo->script, dl, arg, PARSE_STRING_NULL);
+		s = parse_nstring(ifo->script, dl, arg);
 		if (s == -1 ||
 		    ifo->script[0] == '\0' ||
 		    strcmp(ifo->script, "/dev/null") == 0)
@@ -754,7 +760,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			ifo->options |= DHCPCD_HOSTNAME;
 			break;
 		}
-		s = parse_string(ifo->hostname, HOSTNAME_MAX_LEN, arg);
+		s = parse_nstring(ifo->hostname, sizeof(ifo->hostname), arg);
 		if (s == -1) {
 			logerr("%s: hostname", __func__);
 			return -1;
@@ -763,7 +769,6 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			logerrx("hostname cannot begin with .");
 			return -1;
 		}
-		ifo->hostname[s] = '\0';
 		if (ifo->hostname[0] == '\0')
 			ifo->options &= ~DHCPCD_HOSTNAME;
 		else
@@ -874,16 +879,16 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		break;
 	case 't':
 		ARG_REQUIRED;
-		ifo->timeout = (time_t)strtoi(arg, NULL, 0, 0, INT32_MAX, &e);
+		ifo->timeout = (uint32_t)strtou(arg, NULL, 0, 0, UINT32_MAX, &e);
 		if (e) {
 			logerrx("failed to convert timeout %s", arg);
 			return -1;
 		}
 		break;
 	case 'u':
-		s = USERCLASS_MAX_LEN - ifo->userclass[0] - 1;
+		dl = sizeof(ifo->userclass) - ifo->userclass[0] - 1;
 		s = parse_string((char *)ifo->userclass +
-		    ifo->userclass[0] + 2, (size_t)s, arg);
+		    ifo->userclass[0] + 2, dl, arg);
 		if (s == -1) {
 			logerr("userclass");
 			return -1;
@@ -893,6 +898,19 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			ifo->userclass[0] = (uint8_t)(ifo->userclass[0] + s +1);
 		}
 		break;
+#ifndef SMALL
+	case O_MSUSERCLASS:
+		/* Some Microsoft DHCP servers expect userclass to be an
+		 * opaque blob. This is not RFC 3004 compliant. */
+		s = parse_string((char *)ifo->userclass + 1,
+		    sizeof(ifo->userclass) - 1, arg);
+		if (s == -1) {
+			logerr("msuserclass");
+			return -1;
+		}
+		ifo->userclass[0] = (uint8_t)s;
+		break;
+#endif
 	case 'v':
 		ARG_REQUIRED;
 		p = strchr(arg, ',');
@@ -967,7 +985,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		break;
 	case 'y':
 		ARG_REQUIRED;
-		ifo->reboot = (time_t)strtoi(arg, NULL, 0, 0, UINT32_MAX, &e);
+		ifo->reboot = (uint32_t)strtou(arg, NULL, 0, 0, UINT32_MAX, &e);
 		if (e) {
 			logerr("failed to convert reboot %s", arg);
 			return -1;
@@ -1165,7 +1183,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 			np = strchr(p, '/');
 			if (np)
 				*np++ = '\0';
-			if (inet_pton(AF_INET6, p, &ifo->req_addr6) == 1) {
+			if ((i = inet_pton(AF_INET6, p, &ifo->req_addr6)) == 1) {
 				if (np) {
 					ifo->req_prefix_len = (uint8_t)strtou(np,
 					    NULL, 0, 0, 128, &e);
@@ -1178,6 +1196,14 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 				} else
 					ifo->req_prefix_len = 128;
 			}
+			if (np)
+				*(--np) = '\0';
+			if (i != 1) {
+				logerrx("invalid AF_INET6: %s", p);
+				memset(&ifo->req_addr6, 0,
+				    sizeof(ifo->req_addr6));
+				return -1;
+			}
 		} else
 			add_environ(&ifo->config, arg, 1);
 		break;
@@ -1238,6 +1264,34 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 		break;
 	case O_NOIPV6:
 		ifo->options &= ~DHCPCD_IPV6;
+		break;
+	case O_ANONYMOUS:
+		ifo->options |= DHCPCD_ANONYMOUS;
+		ifo->options &= ~DHCPCD_HOSTNAME;
+		ifo->fqdn = FQDN_DISABLE;
+
+		/* Block everything */
+		memset(ifo->nomask, 0xff, sizeof(ifo->nomask));
+		memset(ifo->nomask6, 0xff, sizeof(ifo->nomask6));
+
+		/* Allow the bare minimum through */
+		del_option_mask(ifo->nomask, DHO_SUBNETMASK);
+		del_option_mask(ifo->nomask, DHO_CSR);
+		del_option_mask(ifo->nomask, DHO_ROUTER);
+		del_option_mask(ifo->nomask, DHO_DNSSERVER);
+		del_option_mask(ifo->nomask, DHO_DNSDOMAIN);
+		del_option_mask(ifo->nomask, DHO_BROADCAST);
+		del_option_mask(ifo->nomask, DHO_STATICROUTE);
+		del_option_mask(ifo->nomask, DHO_SERVERID);
+		del_option_mask(ifo->nomask, DHO_RENEWALTIME);
+		del_option_mask(ifo->nomask, DHO_REBINDTIME);
+		del_option_mask(ifo->nomask, DHO_DNSSEARCH);
+
+		del_option_mask(ifo->nomask6, D6_OPTION_DNS_SERVERS);
+		del_option_mask(ifo->nomask6, D6_OPTION_DOMAIN_LIST);
+		del_option_mask(ifo->nomask6, D6_OPTION_SOL_MAX_RT);
+		del_option_mask(ifo->nomask6, D6_OPTION_INF_MAX_RT);
+
 		break;
 #ifdef INET
 	case O_ARPING:
@@ -1398,8 +1452,8 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 				p = strchr(arg, '/');
 				if (p)
 					*p++ = '\0';
-				if (inet_pton(AF_INET6, arg, &ia->addr) == -1) {
-					logerr("%s", arg);
+				if (inet_pton(AF_INET6, arg, &ia->addr) != 1) {
+					logerrx("invalid AF_INET6: %s", arg);
 					memset(&ia->addr, 0, sizeof(ia->addr));
 				}
 				if (p && ia->ia_type == D6_OPTION_IA_PD) {
@@ -1972,52 +2026,42 @@ err_sla:
 			return -1;
 		}
 		*fp++ = '\0';
-		token = malloc(sizeof(*token));
+		token = calloc(1, sizeof(*token));
 		if (token == NULL) {
 			logerr(__func__);
-			free(token);
 			return -1;
 		}
 		if (parse_uint32(&token->secretid, arg) == -1) {
 			logerrx("%s: not a number", arg);
-			free(token);
-			return -1;
+			goto invalid_token;
 		}
 		arg = fp;
 		fp = strend(arg);
 		if (fp == NULL) {
 			logerrx("authtoken requies an a key");
-			free(token);
-			return -1;
+			goto invalid_token;
 		}
 		*fp++ = '\0';
 		s = parse_string(NULL, 0, arg);
 		if (s == -1) {
 			logerr("realm_len");
-			free(token);
-			return -1;
+			goto invalid_token;
 		}
-		if (s) {
+		if (s != 0) {
 			token->realm_len = (size_t)s;
 			token->realm = malloc(token->realm_len);
 			if (token->realm == NULL) {
 				logerr(__func__);
-				free(token);
-				return -1;
+				goto invalid_token;
 			}
 			parse_string((char *)token->realm, token->realm_len,
 			    arg);
-		} else {
-			token->realm_len = 0;
-			token->realm = NULL;
 		}
 		arg = fp;
 		fp = strend(arg);
 		if (fp == NULL) {
 			logerrx("authtoken requies an expiry date");
-			free(token->realm);
-			free(token);
-			return -1;
+			goto invalid_token;
 		}
 		*fp++ = '\0';
 		if (*arg == '"') {
@@ -2034,15 +2078,11 @@ err_sla:
 			memset(&tm, 0, sizeof(tm));
 			if (strptime(arg, "%Y-%m-%d %H:%M", &tm) == NULL) {
 				logerrx("%s: invalid date time", arg);
-				free(token->realm);
-				free(token);
-				return -1;
+				goto invalid_token;
 			}
 			if ((token->expire = mktime(&tm)) == (time_t)-1) {
 				logerr("%s: mktime", __func__);
-				free(token->realm);
-				free(token);
-				return -1;
+				goto invalid_token;
 			}
 		}
 		arg = fp;
@@ -2052,19 +2092,25 @@ err_sla:
 				logerr("token_len");
 			else
 				logerrx("authtoken needs a key");
-			free(token->realm);
-			free(token);
-			return -1;
+			goto invalid_token;
 		}
 		token->key_len = (size_t)s;
 		token->key = malloc(token->key_len);
+		if (token->key == NULL) {
+			logerr(__func__);
+			goto invalid_token;
+		}
 		parse_string((char *)token->key, token->key_len, arg);
 		TAILQ_INSERT_TAIL(&ifo->auth.tokens, token, next);
+		break;
+
+invalid_token:
+		free(token->realm);
+		free(token);
 #else
 		logerrx("no authentication support");
-		return -1;
 #endif
-		break;
+		return -1;
 	case O_AUTHNOTREQUIRED:
 		ifo->auth.options &= ~DHCPCD_AUTH_REQUIRE;
 		break;
@@ -2491,6 +2537,11 @@ read_config(struct dhcpcd_ctx *ctx,
 	}
 
 	/* Parse our options file */
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP &&
+	    ps_root_copychroot(ctx, ctx->cffile) == -1)
+		logwarn("%s: ps_root_copychroot `%s'", __func__, ctx->cffile);
+#endif
 	fp = fopen(ctx->cffile, "r");
 	if (fp == NULL) {
 		/* dhcpcd can continue without it, but no DNS options
