@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * Privilege Separation for dhcpcd
- * Copyright (c) 2006-2021 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -33,14 +33,12 @@
  * or address specific listener.
  * Spawn an unpriv process to send/receive common network data.
  * Then drop all privs and start running.
- * Every process aside from the privileged proxy is chrooted.
- * All privsep processes ignore signals - only the manager process accepts them.
+ * Every process aside from the privileged actioneer is chrooted.
  *
  * dhcpcd will maintain the config file in the chroot, no need to handle
  * this in a script or something.
  */
 
-#include <sys/resource.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -65,7 +63,6 @@
 #include "arp.h"
 #include "common.h"
 #include "control.h"
-#include "dev.h"
 #include "dhcp.h"
 #include "dhcp6.h"
 #include "eloop.h"
@@ -73,28 +70,36 @@
 #include "logerr.h"
 #include "privsep.h"
 
-#ifdef HAVE_CAPSICUM
-#include <sys/capsicum.h>
-#include <capsicum_helpers.h>
-#endif
 #ifdef HAVE_UTIL_H
 #include <util.h>
 #endif
 
-/* CMSG_ALIGN is a Linux extension */
-#ifndef CMSG_ALIGN
-#define CMSG_ALIGN(n)	(CMSG_SPACE((n)) - CMSG_SPACE(0))
-#endif
+int
+ps_mkdir(char *path)
+{
+	char *slash;
+	bool done;
 
-/* Calculate number of padding bytes to achieve 'struct cmsghdr' alignment */
-#define CALC_CMSG_PADLEN(has_cmsg, pos) \
-    ((has_cmsg) ? (socklen_t)(CMSG_ALIGN((pos)) - (pos)) : 0)
+	slash = path;
+	for (;;) {
+		slash += strspn(slash, "/");
+		slash += strcspn(slash, "/");
+		done = (*slash == '\0');
+		*slash = '\0';
+		if (mkdir(path, 0755) == -1 && errno != EEXIST)
+			return -1;
+		if (done)
+			break;
+		*slash = '/';
+	}
+	return 0;
+}
 
 int
 ps_init(struct dhcpcd_ctx *ctx)
 {
-	struct passwd *pw;
-	struct stat st;
+	char path[PATH_MAX];
+	struct passwd *pw = ctx->ps_user;
 
 	errno = 0;
 	if ((ctx->ps_user = pw = getpwnam(PRIVSEP_USER)) == NULL) {
@@ -108,222 +113,52 @@ ps_init(struct dhcpcd_ctx *ctx)
 		return -1;
 	}
 
-	if (stat(pw->pw_dir, &st) == -1 || !S_ISDIR(st.st_mode)) {
-		ctx->options &= ~DHCPCD_PRIVSEP;
-		logerrx("refusing chroot: %s: %s",
-		    PRIVSEP_USER, pw->pw_dir);
+	/* If we pickup the _dhcp user refuse the default directory */
+	if (strcmp(pw->pw_dir, "/var/empty") == 0) {
+		logerrx("refusing chroot: %s: %s", PRIVSEP_USER, pw->pw_dir);
 		errno = 0;
 		return -1;
 	}
 
+	/* Create the database directory. */
+	if (snprintf(path, sizeof(path), "%s%s", pw->pw_dir, DBDIR) == -1 ||
+	    ps_mkdir(path) == -1 ||
+	    chown(path, pw->pw_uid, pw->pw_gid) == -1 ||
+	    chmod(path, 0755) == -1)
+		logerr("%s: %s", __func__, path);
+
+	/* Ensure we have a localtime to correctly format dates. */
+	if (ps_root_docopychroot(ctx, "/etc/localtime") == -1 && errno!=ENOENT)
+		logerr("%s: %s", __func__, "/etc/localtime");
+
 	ctx->options |= DHCPCD_PRIVSEP;
 	return 0;
 }
 
-static int
+int
 ps_dropprivs(struct dhcpcd_ctx *ctx)
 {
 	struct passwd *pw = ctx->ps_user;
 
-	if (ctx->options & DHCPCD_LAUNCHER)
-		logdebugx("chrooting as %s to %s", pw->pw_name, pw->pw_dir);
-	if (chroot(pw->pw_dir) == -1 &&
-	    (errno != EPERM || ctx->options & DHCPCD_FORKED))
-		logerr("%s: chroot: %s", __func__, pw->pw_dir);
-	if (chdir("/") == -1)
-		logerr("%s: chdir: /", __func__);
+	if (!(ctx->options & DHCPCD_FORKED))
+		logdebugx("chrooting to `%s'", pw->pw_dir);
 
-	if ((setgroups(1, &pw->pw_gid) == -1 ||
+	if (chroot(pw->pw_dir) == -1)
+		logerr("%s: chroot `%s'", __func__, pw->pw_dir);
+	if (chdir("/") == -1)
+		logerr("%s: chdir `/'", __func__);
+
+	if (setgroups(1, &pw->pw_gid) == -1 ||
 	     setgid(pw->pw_gid) == -1 ||
-	     setuid(pw->pw_uid) == -1) &&
-	     (errno != EPERM || ctx->options & DHCPCD_FORKED))
+	     setuid(pw->pw_uid) == -1)
 	{
 		logerr("failed to drop privileges");
 		return -1;
 	}
 
-	struct rlimit rzero = { .rlim_cur = 0, .rlim_max = 0 };
-
-	if (ctx->ps_control_pid != getpid()) {
-		/* Prohibit new files, sockets, etc */
-#if defined(__linux__) || defined(__sun) || defined(__OpenBSD__)
-		/*
-		 * If poll(2) is called with nfds > RLIMIT_NOFILE
-		 * then it returns EINVAL.
-		 * This blows.
-		 * Do the best we can and limit to what we need.
-		 * An attacker could potentially close a file and
-		 * open a new one still, but that cannot be helped.
-		 */
-		unsigned long maxfd;
-		maxfd = (unsigned long)eloop_event_count(ctx->eloop);
-		if (IN_PRIVSEP_SE(ctx))
-			maxfd++; /* XXX why? */
-
-		struct rlimit rmaxfd = {
-		    .rlim_cur = maxfd,
-		    .rlim_max = maxfd
-		};
-		if (setrlimit(RLIMIT_NOFILE, &rmaxfd) == -1)
-			logerr("setrlimit RLIMIT_NOFILE");
-#else
-		if (setrlimit(RLIMIT_NOFILE, &rzero) == -1)
-			logerr("setrlimit RLIMIT_NOFILE");
-#endif
-	}
-
-#define DHC_NOCHKIO	(DHCPCD_STARTED | DHCPCD_DAEMONISE)
-	/* Prohibit writing to files.
-	 * Obviously this won't work if we are using a logfile
-	 * or redirecting stderr to a file. */
-	if ((ctx->options & DHC_NOCHKIO) == DHC_NOCHKIO ||
-	    (ctx->logfile == NULL &&
-	    (!ctx->stderr_valid || isatty(STDERR_FILENO) == 1)))
-	{
-		if (setrlimit(RLIMIT_FSIZE, &rzero) == -1)
-			logerr("setrlimit RLIMIT_FSIZE");
-	}
-
-#ifdef RLIMIT_NPROC
-	/* Prohibit forks */
-	if (setrlimit(RLIMIT_NPROC, &rzero) == -1)
-		logerr("setrlimit RLIMIT_NPROC");
-#endif
-
 	return 0;
 }
 
-static int
-ps_setbuf0(int fd, int ctl, int minlen)
-{
-	int len;
-	socklen_t slen;
-
-	slen = sizeof(len);
-	if (getsockopt(fd, SOL_SOCKET, ctl, &len, &slen) == -1)
-		return -1;
-
-#ifdef __linux__
-	len /= 2;
-#endif
-	if (len >= minlen)
-		return 0;
-
-	return setsockopt(fd, SOL_SOCKET, ctl, &minlen, sizeof(minlen));
-}
-
-static int
-ps_setbuf(int fd)
-{
-	/* Ensure we can receive a fully sized privsep message.
-	 * Double the send buffer. */
-	int minlen = (int)sizeof(struct ps_msg);
-
-	if (ps_setbuf0(fd, SO_RCVBUF, minlen) == -1 ||
-	    ps_setbuf0(fd, SO_SNDBUF, minlen * 2) == -1)
-	{
-		logerr(__func__);
-		return -1;
-	}
-	return 0;
-}
-
-int
-ps_setbuf_fdpair(int fd[])
-{
-
-	if (ps_setbuf(fd[0]) == -1 || ps_setbuf(fd[1]) == -1)
-		return -1;
-	return 0;
-}
-
-#ifdef PRIVSEP_RIGHTS
-int
-ps_rights_limit_ioctl(int fd)
-{
-	cap_rights_t rights;
-
-	cap_rights_init(&rights, CAP_IOCTL);
-	if (cap_rights_limit(fd, &rights) == -1 && errno != ENOSYS)
-		return -1;
-	return 0;
-}
-
-int
-ps_rights_limit_fd_fctnl(int fd)
-{
-	cap_rights_t rights;
-
-	cap_rights_init(&rights, CAP_READ, CAP_WRITE, CAP_EVENT,
-	    CAP_ACCEPT, CAP_FCNTL);
-	if (cap_rights_limit(fd, &rights) == -1 && errno != ENOSYS)
-		return -1;
-	return 0;
-}
-
-int
-ps_rights_limit_fd(int fd)
-{
-	cap_rights_t rights;
-
-	cap_rights_init(&rights, CAP_READ, CAP_WRITE, CAP_EVENT, CAP_SHUTDOWN);
-	if (cap_rights_limit(fd, &rights) == -1 && errno != ENOSYS)
-		return -1;
-	return 0;
-}
-
-int
-ps_rights_limit_fd_sockopt(int fd)
-{
-	cap_rights_t rights;
-
-	cap_rights_init(&rights, CAP_READ, CAP_WRITE, CAP_EVENT,
-	    CAP_GETSOCKOPT, CAP_SETSOCKOPT);
-	if (cap_rights_limit(fd, &rights) == -1 && errno != ENOSYS)
-		return -1;
-	return 0;
-}
-
-int
-ps_rights_limit_fd_rdonly(int fd)
-{
-	cap_rights_t rights;
-
-	cap_rights_init(&rights, CAP_READ, CAP_EVENT);
-	if (cap_rights_limit(fd, &rights) == -1 && errno != ENOSYS)
-		return -1;
-	return 0;
-}
-
-int
-ps_rights_limit_fdpair(int fd[])
-{
-
-	if (ps_rights_limit_fd(fd[0]) == -1 || ps_rights_limit_fd(fd[1]) == -1)
-		return -1;
-	return 0;
-}
-
-static int
-ps_rights_limit_stdio(struct dhcpcd_ctx *ctx)
-{
-	const int iebadf = CAPH_IGNORE_EBADF;
-	int error = 0;
-
-	if (ctx->stdin_valid &&
-	    caph_limit_stream(STDIN_FILENO, CAPH_READ | iebadf) == -1)
-		error = -1;
-	if (ctx->stdout_valid &&
-	    caph_limit_stream(STDOUT_FILENO, CAPH_WRITE | iebadf) == -1)
-		error = -1;
-	if (ctx->stderr_valid &&
-	    caph_limit_stream(STDERR_FILENO, CAPH_WRITE | iebadf) == -1)
-		error = -1;
-
-	return error;
-}
-#endif
-
 pid_t
 ps_dostart(struct dhcpcd_ctx *ctx,
     pid_t *priv_pid, int *priv_fd,
@@ -331,23 +166,15 @@ ps_dostart(struct dhcpcd_ctx *ctx,
     void *recv_ctx, int (*callback)(void *), void (*signal_cb)(int, void *),
     unsigned int flags)
 {
+	int stype;
 	int fd[2];
 	pid_t pid;
 
-	if (xsocketpair(AF_UNIX, SOCK_DGRAM | SOCK_CXNB, 0, fd) == -1) {
-		logerr("%s: socketpair", __func__);
+	stype = SOCK_CLOEXEC | SOCK_NONBLOCK;
+	if (socketpair(AF_UNIX, SOCK_DGRAM | stype, 0, fd) == -1) {
+		logerr("socketpair");
 		return -1;
 	}
-	if (ps_setbuf_fdpair(fd) == -1) {
-		logerr("%s: ps_setbuf_fdpair", __func__);
-		return -1;
-	}
-#ifdef PRIVSEP_RIGHTS
-	if (ps_rights_limit_fdpair(fd) == -1) {
-		logerr("%s: ps_rights_limit_fdpair", __func__);
-		return -1;
-	}
-#endif
 
 	switch (pid = fork()) {
 	case -1:
@@ -361,9 +188,8 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 		*priv_pid = pid;
 		*priv_fd = fd[0];
 		close(fd[1]);
-		if (recv_unpriv_msg == NULL)
-			;
-		else if (eloop_event_add(ctx->eloop, *priv_fd,
+		if (recv_unpriv_msg != NULL &&
+		    eloop_event_add(ctx->eloop, *priv_fd,
 		    recv_unpriv_msg, recv_ctx) == -1)
 		{
 			logerr("%s: eloop_event_add", __func__);
@@ -372,21 +198,13 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 		return pid;
 	}
 
-	ctx->options |= DHCPCD_FORKED;
+	ctx->options |= DHCPCD_UNPRIV | DHCPCD_FORKED;
 	if (ctx->fork_fd != -1) {
 		close(ctx->fork_fd);
 		ctx->fork_fd = -1;
 	}
 	pidfile_clean();
-
 	eloop_clear(ctx->eloop);
-	eloop_signal_set_cb(ctx->eloop,
-	    dhcpcd_signals, dhcpcd_signals_len, signal_cb, ctx);
-	/* ctx->sigset aready has the initial sigmask set in main() */
-	if (eloop_signal_mask(ctx->eloop, NULL) == -1) {
-		logerr("%s: eloop_signal_mask", __func__);
-		goto errexit;
-	}
 
 	/* We are not root */
 	if (priv_fd != &ctx->ps_root_fd) {
@@ -395,14 +213,6 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 			close(ctx->ps_root_fd);
 			ctx->ps_root_fd = -1;
 		}
-
-#ifdef PRIVSEP_RIGHTS
-		/* We cannot limit the root process in any way. */
-		if (ps_rights_limit_stdio(ctx) == -1) {
-			logerr("ps_rights_limit_stdio");
-			goto errexit;
-		}
-#endif
 	}
 
 	if (priv_fd != &ctx->ps_inet_fd && ctx->ps_inet_fd != -1) {
@@ -410,6 +220,17 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 		ctx->ps_inet_fd = -1;
 	}
 
+	if (eloop_signal_set_cb(ctx->eloop,
+	    dhcpcd_signals, dhcpcd_signals_len, signal_cb, ctx) == -1)
+	{
+		logerr("%s: eloop_signal_set_cb", __func__);
+		goto errexit;
+	}
+	if (eloop_signal_mask(ctx->eloop, &ctx->sigset) == -1) {
+		logerr("%s: eloop_signal_mask", __func__);
+		goto errexit;
+	}
+
 	if (eloop_event_add(ctx->eloop, *priv_fd, recv_msg, recv_ctx) == -1)
 	{
 		logerr("%s: eloop_event_add", __func__);
@@ -419,6 +240,11 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 	if (callback(recv_ctx) == -1)
 		goto errexit;
 
+	if (!(ctx->options & (DHCPCD_TEST | DHCPCD_DEBUG))) {
+		freopen(_PATH_DEVNULL, "w", stdout);
+		freopen(_PATH_DEVNULL, "w", stderr);
+	}
+
 	if (flags & PSF_DROPPRIVS)
 		ps_dropprivs(ctx);
 
@@ -427,37 +253,55 @@ ps_dostart(struct dhcpcd_ctx *ctx,
 errexit:
 	/* Failure to start root or inet processes is fatal. */
 	if (priv_fd == &ctx->ps_root_fd || priv_fd == &ctx->ps_inet_fd)
-		(void)ps_sendcmd(ctx, *priv_fd, PS_STOP, 0, NULL, 0);
+		ps_sendcmd(ctx, *priv_fd, PS_STOP, 0, NULL, 0);
 	shutdown(*priv_fd, SHUT_RDWR);
 	*priv_fd = -1;
-	eloop_exit(ctx->eloop, EXIT_FAILURE);
 	return -1;
 }
 
 int
 ps_dostop(struct dhcpcd_ctx *ctx, pid_t *pid, int *fd)
 {
-	int err = 0;
+	int status;
 
 #ifdef PRIVSEP_DEBUG
-	logdebugx("%s: pid=%d fd=%d", __func__, *pid, *fd);
+	logdebugx("%s: pid %d fd %d", __func__, *pid, *fd);
+#endif
+	if (*pid == 0)
+		return 0;
+	eloop_event_delete(ctx->eloop, *fd);
+	if (ps_sendcmd(ctx, *fd, PS_STOP, 0, NULL, 0) == -1 &&
+	    errno != ECONNRESET)
+		logerr(__func__);
+	if (shutdown(*fd, SHUT_RDWR) == -1 && errno != ENOTCONN)
+		logerr(__func__);
+	close(*fd);
+	*fd = -1;
+	/* We won't have permission for all processes .... */
+#if 0
+	if (kill(*pid, SIGTERM) == -1)
+		logerr(__func__);
+#endif
+	status = 0;
+	/* Wait for the process to finish */
+	while (waitpid(*pid, &status, 0) == -1) {
+		if (errno != EINTR) {
+			logerr("%s: waitpid", __func__);
+			status = 0;
+			break;
+		}
+#ifdef PRIVSEP_DEBUG
+		else
+			logerr("%s: waitpid ", __func__);
+#endif
+	}
+	*pid = 0;
+
+#ifdef PRIVSEP_DEBUG
+	logdebugx("%s: status %d", __func__, status);
 #endif
 
-	if (*fd != -1) {
-		eloop_event_delete(ctx->eloop, *fd);
-		if (ps_sendcmd(ctx, *fd, PS_STOP, 0, NULL, 0) == -1) {
-			logerr(__func__);
-			err = -1;
-		}
-		(void)shutdown(*fd, SHUT_RDWR);
-		close(*fd);
-		*fd = -1;
-	}
-
-	/* Don't wait for the process as it may not respond to the shutdown
-	 * request. We'll reap the process on receipt of SIGCHLD. */
-	*pid = 0;
-	return err;
+	return status;
 }
 
 int
@@ -469,21 +313,22 @@ ps_start(struct dhcpcd_ctx *ctx)
 
 	switch (pid = ps_root_start(ctx)) {
 	case -1:
-		logerr("ps_root_start");
 		return -1;
 	case 0:
 		return 0;
 	default:
-		logdebugx("spawned privileged proxy on PID %d", pid);
+		logdebugx("spawned privileged actioneer on PID %d", pid);
 	}
 
 	/* No point in spawning the generic network listener if we're
 	 * not going to use it. */
-	if (!ps_inet_canstart(ctx))
-		goto started_net;
+	if (!(ctx->options & (DHCPCD_MASTER | DHCPCD_IPV6RS)))
+		goto started;
 
 	switch (pid = ps_inet_start(ctx)) {
 	case -1:
+		if (errno == ENXIO)
+			return 0;
 		return -1;
 	case 0:
 		return 0;
@@ -491,111 +336,10 @@ ps_start(struct dhcpcd_ctx *ctx)
 		logdebugx("spawned network proxy on PID %d", pid);
 	}
 
-started_net:
-	if (!(ctx->options & DHCPCD_TEST)) {
-		switch (pid = ps_ctl_start(ctx)) {
-		case -1:
-			return -1;
-		case 0:
-			return 0;
-		default:
-			logdebugx("spawned controller proxy on PID %d", pid);
-		}
-	}
-
-#ifdef ARC4RANDOM_H
-	/* Seed the random number generator early incase it needs /dev/urandom
-	 * which won't be available in the chroot. */
-	arc4random();
-#endif
-
+started:
 	return 1;
 }
 
-int
-ps_entersandbox(const char *_pledge, const char **sandbox)
-{
-
-#if !defined(HAVE_PLEDGE)
-	UNUSED(_pledge);
-#endif
-
-#if defined(HAVE_CAPSICUM)
-	if (sandbox != NULL)
-		*sandbox = "capsicum";
-	return cap_enter();
-#elif defined(HAVE_PLEDGE)
-	if (sandbox != NULL)
-		*sandbox = "pledge";
-	return pledge(_pledge, NULL);
-#elif defined(HAVE_SECCOMP)
-	if (sandbox != NULL)
-		*sandbox = "seccomp";
-	return ps_seccomp_enter();
-#else
-	if (sandbox != NULL)
-		*sandbox = "posix resource limited";
-	return 0;
-#endif
-}
-
-int
-ps_managersandbox(struct dhcpcd_ctx *ctx, const char *_pledge)
-{
-	const char *sandbox = NULL;
-	bool forked;
-	int dropped;
-
-	forked = ctx->options & DHCPCD_FORKED;
-	ctx->options &= ~DHCPCD_FORKED;
-	dropped = ps_dropprivs(ctx);
-	if (forked)
-		ctx->options |= DHCPCD_FORKED;
-
-	/*
-	 * If we don't have a root process, we cannot use syslog.
-	 * If it cannot be opened before chrooting then syslog(3) will fail.
-	 * openlog(3) does not return an error which doubly sucks.
-	 */
-	if (ctx->ps_root_fd == -1) {
-		unsigned int logopts = loggetopts();
-
-		logopts &= ~LOGERR_LOG;
-		logsetopts(logopts);
-	}
-
-	if (dropped == -1) {
-		logerr("%s: ps_dropprivs", __func__);
-		return -1;
-	}
-
-#ifdef PRIVSEP_RIGHTS
-	if ((ctx->pf_inet_fd != -1 &&
-	    ps_rights_limit_ioctl(ctx->pf_inet_fd) == -1) ||
-	     ps_rights_limit_stdio(ctx) == -1)
-	{
-		logerr("%s: cap_rights_limit", __func__);
-		return -1;
-	}
-#endif
-
-	if (_pledge == NULL)
-		_pledge = "stdio";
-	if (ps_entersandbox(_pledge, &sandbox) == -1) {
-		if (errno == ENOSYS) {
-			if (sandbox != NULL)
-				logwarnx("sandbox unavailable: %s", sandbox);
-			return 0;
-		}
-		logerr("%s: %s", __func__, sandbox);
-		return -1;
-	} else if (ctx->options & DHCPCD_LAUNCHER ||
-		  ((!(ctx->options & DHCPCD_DAEMONISE)) &&
-		   ctx->options & DHCPCD_MANAGER))
-		logdebugx("sandbox: %s", sandbox);
-	return 0;
-}
-
 int
 ps_stop(struct dhcpcd_ctx *ctx)
 {
@@ -606,16 +350,12 @@ ps_stop(struct dhcpcd_ctx *ctx)
 	    ctx->eloop == NULL)
 		return 0;
 
-	r = ps_ctl_stop(ctx);
-	if (r != 0)
-		ret = r;
-
 	r = ps_inet_stop(ctx);
 	if (r != 0)
 		ret = r;
 
 	/* We've been chrooted, so we need to tell the
-	 * privileged proxy to remove the pidfile. */
+	 * privileged actioneer to remove the pidfile. */
 	ps_root_unlink(ctx, ctx->pidfile);
 
 	r = ps_root_stop(ctx);
@@ -629,6 +369,14 @@ ps_stop(struct dhcpcd_ctx *ctx)
 void
 ps_freeprocess(struct ps_process *psp)
 {
+#ifdef INET
+	struct ipv4_state *istate = IPV4_STATE(&psp->psp_ifp);
+
+	if (istate != NULL) {
+		free(istate->buffer);
+		free(istate);
+	}
+#endif
 
 	TAILQ_REMOVE(&psp->psp_ctx->ps_processes, psp, next);
 	if (psp->psp_fd != -1) {
@@ -639,10 +387,6 @@ ps_freeprocess(struct ps_process *psp)
 		eloop_event_delete(psp->psp_ctx->eloop, psp->psp_work_fd);
 		close(psp->psp_work_fd);
 	}
-#ifdef INET
-	if (psp->psp_bpf != NULL)
-		bpf_close(psp->psp_bpf);
-#endif
 	free(psp);
 }
 
@@ -664,11 +408,9 @@ ps_unrollmsg(struct msghdr *msg, struct ps_msghdr *psm,
     const void *data, size_t len)
 {
 	uint8_t *datap, *namep, *controlp;
-	socklen_t cmsg_padlen =
-	    CALC_CMSG_PADLEN(psm->ps_controllen, psm->ps_namelen);
 
 	namep = UNCONST(data);
-	controlp = namep + psm->ps_namelen + cmsg_padlen;
+	controlp = namep + psm->ps_namelen;
 	datap = controlp + psm->ps_controllen;
 
 	if (psm->ps_namelen != 0) {
@@ -688,7 +430,7 @@ ps_unrollmsg(struct msghdr *msg, struct ps_msghdr *psm,
 			return -1;
 		}
 		msg->msg_control = controlp;
-		len -= psm->ps_controllen + cmsg_padlen;
+		len -= psm->ps_controllen;
 	} else
 		msg->msg_control = NULL;
 	msg->msg_controllen = psm->ps_controllen;
@@ -705,27 +447,26 @@ ps_unrollmsg(struct msghdr *msg, struct ps_msghdr *psm,
 	return 0;
 }
 
+
 ssize_t
 ps_sendpsmmsg(struct dhcpcd_ctx *ctx, int fd,
     struct ps_msghdr *psm, const struct msghdr *msg)
 {
-	long padding[1] = { 0 };
+	assert(msg == NULL || msg->msg_iovlen == 1);
+
 	struct iovec iov[] = {
 		{ .iov_base = UNCONST(psm), .iov_len = sizeof(*psm) },
 		{ .iov_base = NULL, },	/* name */
-		{ .iov_base = NULL, },	/* control padding */
 		{ .iov_base = NULL, },	/* control */
-		{ .iov_base = NULL, },	/* payload 1 */
-		{ .iov_base = NULL, },	/* payload 2 */
-		{ .iov_base = NULL, },	/* payload 3 */
+		{ .iov_base = NULL, },	/* payload */
 	};
-	int iovlen;
+	int iovlen = __arraycount(iov);
 	ssize_t len;
 
 	if (msg != NULL) {
 		struct iovec *iovp = &iov[1];
-		int i;
-		socklen_t cmsg_padlen;
+
+		assert(msg->msg_iovlen == 1);
 
 		psm->ps_namelen = msg->msg_namelen;
 		psm->ps_controllen = (socklen_t)msg->msg_controllen;
@@ -733,38 +474,21 @@ ps_sendpsmmsg(struct dhcpcd_ctx *ctx, int fd,
 		iovp->iov_base = msg->msg_name;
 		iovp->iov_len = msg->msg_namelen;
 		iovp++;
-
-		cmsg_padlen =
-		    CALC_CMSG_PADLEN(msg->msg_controllen, msg->msg_namelen);
-		assert(cmsg_padlen <= sizeof(padding));
-		iovp->iov_len = cmsg_padlen;
-		iovp->iov_base = cmsg_padlen != 0 ? padding : NULL;
-		iovp++;
-
 		iovp->iov_base = msg->msg_control;
 		iovp->iov_len = msg->msg_controllen;
-		iovlen = 4;
-
-		for (i = 0; i < (int)msg->msg_iovlen; i++) {
-			if ((size_t)(iovlen + i) > __arraycount(iov)) {
-				errno =	ENOBUFS;
-				return -1;
-			}
-			iovp++;
-			iovp->iov_base = msg->msg_iov[i].iov_base;
-			iovp->iov_len = msg->msg_iov[i].iov_len;
-		}
-		iovlen += i;
+		iovp++;
+		iovp->iov_base = msg->msg_iov[0].iov_base;
+		iovp->iov_len = msg->msg_iov[0].iov_len;
+		iovlen = __arraycount(iov);
 	} else
 		iovlen = 1;
 
 	len = writev(fd, iov, iovlen);
-	if (len == -1) {
-		logerr(__func__);
-		if (ctx->options & DHCPCD_FORKED &&
-		    !(ctx->options & DHCPCD_PRIVSEPROOT))
-			eloop_exit(ctx->eloop, EXIT_FAILURE);
-	}
+#ifdef PRIVSEP_DEBUG
+	logdebugx("%s: %zd", __func__, len);
+#endif
+	if ((len == -1 || len == 0) && ctx->options & DHCPCD_FORKED)
+		eloop_exit(ctx->eloop, len == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 	return len;
 }
 
@@ -784,19 +508,18 @@ ps_sendpsmdata(struct dhcpcd_ctx *ctx, int fd,
 
 
 ssize_t
-ps_sendmsg(struct dhcpcd_ctx *ctx, int fd, uint16_t cmd, unsigned long flags,
+ps_sendmsg(struct dhcpcd_ctx *ctx, int fd, uint8_t cmd, unsigned long flags,
     const struct msghdr *msg)
 {
+	assert(msg->msg_iovlen == 1);
+
 	struct ps_msghdr psm = {
 		.ps_cmd = cmd,
 		.ps_flags = flags,
 		.ps_namelen = msg->msg_namelen,
 		.ps_controllen = (socklen_t)msg->msg_controllen,
+		.ps_datalen = msg->msg_iov[0].iov_len,
 	};
-	size_t i;
-
-	for (i = 0; i < (size_t)msg->msg_iovlen; i++)
-		psm.ps_datalen += msg->msg_iov[i].iov_len;
 
 #if 0	/* For debugging structure padding. */
 	logerrx("psa.family %lu %zu", offsetof(struct ps_addr, psa_family), sizeof(psm.ps_id.psi_addr.psa_family));
@@ -827,7 +550,7 @@ ps_sendmsg(struct dhcpcd_ctx *ctx, int fd, uint16_t cmd, unsigned long flags,
 }
 
 ssize_t
-ps_sendcmd(struct dhcpcd_ctx *ctx, int fd, uint16_t cmd, unsigned long flags,
+ps_sendcmd(struct dhcpcd_ctx *ctx, int fd, uint8_t cmd, unsigned long flags,
     const void *data, size_t len)
 {
 	struct ps_msghdr psm = {
@@ -845,7 +568,7 @@ ps_sendcmd(struct dhcpcd_ctx *ctx, int fd, uint16_t cmd, unsigned long flags,
 }
 
 static ssize_t
-ps_sendcmdmsg(int fd, uint16_t cmd, const struct msghdr *msg)
+ps_sendcmdmsg(int fd, uint8_t cmd, const struct msghdr *msg)
 {
 	struct ps_msghdr psm = { .ps_cmd = cmd };
 	uint8_t data[PS_BUFLEN], *p = data;
@@ -854,8 +577,6 @@ ps_sendcmdmsg(int fd, uint16_t cmd, const struct msghdr *msg)
 		{ .iov_base = data, .iov_len = 0 },
 	};
 	size_t dl = sizeof(data);
-	socklen_t cmsg_padlen =
-	    CALC_CMSG_PADLEN(msg->msg_controllen, msg->msg_namelen);
 
 	if (msg->msg_namelen != 0) {
 		if (msg->msg_namelen > dl)
@@ -867,13 +588,8 @@ ps_sendcmdmsg(int fd, uint16_t cmd, const struct msghdr *msg)
 	}
 
 	if (msg->msg_controllen != 0) {
-		if (msg->msg_controllen + cmsg_padlen > dl)
+		if (msg->msg_controllen > dl)
 			goto nobufs;
-		if (cmsg_padlen != 0) {
-			memset(p, 0, cmsg_padlen);
-			p += cmsg_padlen;
-			dl -= cmsg_padlen;
-		}
 		psm.ps_controllen = (socklen_t)msg->msg_controllen;
 		memcpy(p, msg->msg_control, msg->msg_controllen);
 		p += msg->msg_controllen;
@@ -884,8 +600,7 @@ ps_sendcmdmsg(int fd, uint16_t cmd, const struct msghdr *msg)
 	if (psm.ps_datalen > dl)
 		goto nobufs;
 
-	iov[1].iov_len =
-	    psm.ps_namelen + psm.ps_controllen + psm.ps_datalen + cmsg_padlen;
+	iov[1].iov_len = psm.ps_namelen + psm.ps_controllen + psm.ps_datalen;
 	if (psm.ps_datalen != 0)
 		memcpy(p, msg->msg_iov[0].iov_base, psm.ps_datalen);
 	return writev(fd, iov, __arraycount(iov));
@@ -896,7 +611,7 @@ nobufs:
 }
 
 ssize_t
-ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, uint16_t cmd, int wfd)
+ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, uint8_t cmd, int wfd)
 {
 	struct sockaddr_storage ss = { .ss_family = AF_UNSPEC };
 	uint8_t controlbuf[sizeof(struct sockaddr_storage)] = { 0 };
@@ -911,25 +626,21 @@ ps_recvmsg(struct dhcpcd_ctx *ctx, int rfd, uint16_t cmd, int wfd)
 	};
 
 	ssize_t len = recvmsg(rfd, &msg, 0);
-
-	if (len == -1)
-		logerr("%s: recvmsg", __func__);
-	if (len == -1 || len == 0) {
-		if (ctx->options & DHCPCD_FORKED &&
-		    !(ctx->options & DHCPCD_PRIVSEPROOT))
-			eloop_exit(ctx->eloop,
-			    len == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
+#ifdef PRIVSEP_DEBUG
+	logdebugx("%s: recv fd %d, %zd bytes", __func__, rfd, len);
+#endif
+	if ((len == -1 || len == 0) && ctx->options & DHCPCD_FORKED) {
+		eloop_exit(ctx->eloop, len == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 		return len;
 	}
 
 	iov[0].iov_len = (size_t)len;
 	len = ps_sendcmdmsg(wfd, cmd, &msg);
-	if (len == -1) {
-		logerr("ps_sendcmdmsg");
-		if (ctx->options & DHCPCD_FORKED &&
-		    !(ctx->options & DHCPCD_PRIVSEPROOT))
-			eloop_exit(ctx->eloop, EXIT_FAILURE);
-	}
+#ifdef PRIVSEP_DEBUG
+	logdebugx("%s: send fd %d, %zu bytes", __func__, wfd, len);
+#endif
+	if ((len == -1 || len == 0) && ctx->options & DHCPCD_FORKED)
+		eloop_exit(ctx->eloop, len == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
 	return len;
 }
 
@@ -950,6 +661,8 @@ ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
 	logdebugx("%s: %zd", __func__, len);
 #endif
 
+	if (len == -1 && (errno == ECONNRESET || errno == EBADF))
+		len = 0;
 	if (len == -1 || len == 0)
 		stop = true;
 	else {
@@ -970,9 +683,6 @@ ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
 		logdebugx("process %d stopping", getpid());
 #endif
 		ps_free(ctx);
-#ifdef PLUGIN_DEV
-		dev_stop(ctx);
-#endif
 		eloop_exit(ctx->eloop, len != -1 ? EXIT_SUCCESS : EXIT_FAILURE);
 		return len;
 	}
@@ -981,9 +691,6 @@ ps_recvpsmsg(struct dhcpcd_ctx *ctx, int fd,
 	if (ps_unrollmsg(&msg, &psm.psm_hdr, psm.psm_data, dlen) == -1)
 		return -1;
 
-	if (callback == NULL)
-		return 0;
-
 	errno = 0;
 	return callback(cbctx, &psm.psm_hdr, &msg);
 }
