@@ -1,7 +1,7 @@
-/* SPDX-License-Identifier: BSD-2-Clause */
+/* stSPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2019 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -36,6 +36,7 @@
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
+#include <pwd.h>
 #include <signal.h>
 #include <spawn.h>
 #include <stdarg.h>
@@ -47,18 +48,15 @@
 #include "common.h"
 #include "dhcp.h"
 #include "dhcp6.h"
+#include "eloop.h"
 #include "if.h"
 #include "if-options.h"
 #include "ipv4ll.h"
 #include "ipv6nd.h"
 #include "logerr.h"
+#include "privsep.h"
 #include "script.h"
 
-/* Allow the OS to define another script env var name */
-#ifndef RC_SVCNAME
-#define RC_SVCNAME "RC_SVCNAME"
-#endif
-
 #define DEFAULT_PATH	"/usr/bin:/usr/sbin:/bin:/sbin"
 
 static const char * const if_params[] = {
@@ -85,7 +83,7 @@ if_printoptions(void)
 		printf(" -  %s\n", *p);
 }
 
-static int
+pid_t
 script_exec(const struct dhcpcd_ctx *ctx, char *const *argv, char *const *env)
 {
 	pid_t pid = 0;
@@ -173,7 +171,7 @@ efprintf(FILE *fp, const char *fmt, ...)
 	return r;
 }
 
-static char **
+char **
 script_buftoenv(struct dhcpcd_ctx *ctx, char *buf, size_t len)
 {
 	char **env, **envp, *bufp, *endp;
@@ -216,14 +214,14 @@ script_buftoenv(struct dhcpcd_ctx *ctx, char *buf, size_t len)
 }
 
 static long
-make_env(const struct interface *ifp, const char *reason)
+make_env(struct dhcpcd_ctx *ctx, const struct interface *ifp,
+    const char *reason)
 {
-	struct dhcpcd_ctx *ctx = ifp->ctx;
 	FILE *fp;
 	long buf_pos, i;
 	char *path;
 	int protocol = PROTO_LINK;
-	const struct if_options *ifo = ifp->options;
+	const struct if_options *ifo;
 	const struct interface *ifp2;
 	int af;
 #ifdef INET
@@ -262,6 +260,25 @@ make_env(const struct interface *ifp, const char *reason)
 	}
 #endif
 
+	/* Needed for scripts */
+	path = getenv("PATH");
+	if (efprintf(fp, "PATH=%s", path == NULL ? DEFAULT_PATH:path) == -1)
+		goto eexit;
+	if (efprintf(fp, "reason=%s", reason) == -1)
+		goto eexit;
+	if (efprintf(fp, "pid=%d", getpid()) == -1)
+		goto eexit;
+
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP && ctx->ps_user != NULL) {
+		if (efprintf(fp, "chroot=%s", ctx->ps_user->pw_dir) == -1)
+			goto eexit;
+	}
+	if (strcmp(reason, "CHROOT") == 0)
+		goto make;
+#endif
+
+	ifo = ifp->options;
 #ifdef INET
 	state = D_STATE(ifp);
 #ifdef IPV4LL
@@ -272,7 +289,10 @@ make_env(const struct interface *ifp, const char *reason)
 	d6_state = D6_CSTATE(ifp);
 #endif
 	if (strcmp(reason, "TEST") == 0) {
-		if (1 == 2) {}
+		if (1 == 2) {
+			/* This space left intentionally blank
+			 * as all the below statements are optional. */
+		}
 #ifdef INET6
 #ifdef DHCP6
 		else if (d6_state && d6_state->new)
@@ -316,19 +336,11 @@ make_env(const struct interface *ifp, const char *reason)
 		protocol = PROTO_DHCP;
 #endif
 
-	/* Needed for scripts */
-	path = getenv("PATH");
-	if (efprintf(fp, "PATH=%s", path == NULL ? DEFAULT_PATH:path) == -1)
-		goto eexit;
 
 	if (efprintf(fp, "interface=%s", ifp->name) == -1)
 		goto eexit;
-	if (efprintf(fp, "reason=%s", reason) == -1)
-		goto eexit;
 	if (ifp->ctx->options & DHCPCD_DUMPLEASE)
 		goto dumplease;
-	if (efprintf(fp, "pid=%d", getpid()) == -1)
-		goto eexit;
 	if (efprintf(fp, "ifcarrier=%s",
 	    ifp->carrier == LINK_UNKNOWN ? "unknown" :
 	    ifp->carrier == LINK_UP ? "up" : "down") == -1)
@@ -453,7 +465,7 @@ make_env(const struct interface *ifp, const char *reason)
 dumplease:
 #ifdef INET
 #ifdef IPV4LL
-	if (protocol == PROTO_IPV4LL) {
+	if (protocol == PROTO_IPV4LL && istate) {
 		if (ipv4ll_env(fp, istate->down ? "old" : "new", ifp) == -1)
 			goto eexit;
 	}
@@ -492,6 +504,9 @@ dumplease:
 				goto eexit;
 	}
 
+#ifdef PRIVSEP
+make:
+#endif
 	/* Convert buffer to argv */
 	fflush(fp);
 
@@ -520,7 +535,7 @@ dumplease:
 	if (script_buftoenv(ctx, ctx->script_buf, (size_t)buf_pos) == NULL)
 		goto eexit;
 
-	return buf_pos - 1;
+	return buf_pos;
 
 eexit:
 	logerr(__func__);
@@ -538,16 +553,15 @@ send_interface1(struct fd_list *fd, const struct interface *ifp,
 	struct dhcpcd_ctx *ctx = ifp->ctx;
 	long len;
 
-	len = make_env(ifp, reason);
+	len = make_env(ifp->ctx, ifp, reason);
 	if (len == -1)
 		return -1;
-	return control_queue(fd, ctx->script_buf,  (size_t)len, 1);
+	return control_queue(fd, ctx->script_buf, (size_t)len, 1);
 }
 
 int
-send_interface(struct fd_list *fd, const struct interface *ifp)
+send_interface(struct fd_list *fd, const struct interface *ifp, int af)
 {
-	const char *reason;
 	int retval = 0;
 #ifdef INET
 	const struct dhcp_state *d;
@@ -556,80 +570,92 @@ send_interface(struct fd_list *fd, const struct interface *ifp)
 	const struct dhcp6_state *d6;
 #endif
 
-	switch (ifp->carrier) {
-	case LINK_UP:
-		reason = "CARRIER";
-		break;
-	case LINK_DOWN:
-	case LINK_DOWN_IFFUP:
-		reason = "NOCARRIER";
-		break;
-	default:
-		reason = "UNKNOWN";
-		break;
-	}
-	if (send_interface1(fd, ifp, reason) == -1)
-		retval = -1;
-#ifdef INET
-	if (D_STATE_RUNNING(ifp)) {
-		d = D_CSTATE(ifp);
-		if (send_interface1(fd, ifp, d->reason) == -1)
-			retval = -1;
-	}
-#ifdef IPV4LL
-	if (IPV4LL_STATE_RUNNING(ifp)) {
-		if (send_interface1(fd, ifp, "IPV4LL") == -1)
-			retval = -1;
-	}
+#ifndef AF_LINK
+#define	AF_LINK	AF_PACKET
 #endif
+
+	if (af == AF_UNSPEC || af == AF_LINK) {
+		const char *reason;
+
+		switch (ifp->carrier) {
+		case LINK_UP:
+			reason = "CARRIER";
+			break;
+		case LINK_DOWN:
+		case LINK_DOWN_IFFUP:
+			reason = "NOCARRIER";
+			break;
+		default:
+			reason = "UNKNOWN";
+			break;
+		}
+		if (fd != NULL) {
+			if (send_interface1(fd, ifp, reason) == -1)
+				retval = -1;
+		} else
+			retval++;
+	}
+
+#ifdef INET
+	if (af == AF_UNSPEC || af == AF_INET) {
+		if (D_STATE_RUNNING(ifp)) {
+			d = D_CSTATE(ifp);
+			if (fd != NULL) {
+				if (send_interface1(fd, ifp, d->reason) == -1)
+					retval = -1;
+			} else
+				retval++;
+		}
+#ifdef IPV4LL
+		if (IPV4LL_STATE_RUNNING(ifp)) {
+			if (fd != NULL) {
+				if (send_interface1(fd, ifp, "IPV4LL") == -1)
+					retval = -1;
+			} else
+				retval++;
+		}
+#endif
+	}
 #endif
 
 #ifdef INET6
-	if (IPV6_STATE_RUNNING(ifp)) {
-		if (send_interface1(fd, ifp, "STATIC6") == -1)
-			retval = -1;
-	}
-	if (RS_STATE_RUNNING(ifp)) {
-		if (send_interface1(fd, ifp, "ROUTERADVERT") == -1)
-			retval = -1;
-	}
+	if (af == AF_UNSPEC || af == AF_INET6) {
+		if (IPV6_STATE_RUNNING(ifp)) {
+			if (fd != NULL) {
+				if (send_interface1(fd, ifp, "STATIC6") == -1)
+					retval = -1;
+			} else
+				retval++;
+		}
+		if (RS_STATE_RUNNING(ifp)) {
+			if (fd != NULL) {
+				if (send_interface1(fd, ifp,
+				    "ROUTERADVERT") == -1)
+					retval = -1;
+			} else
+				retval++;
+		}
 #ifdef DHCP6
-	if (D6_STATE_RUNNING(ifp)) {
-		d6 = D6_CSTATE(ifp);
-		if (send_interface1(fd, ifp, d6->reason) == -1)
-			retval = -1;
-	}
+		if (D6_STATE_RUNNING(ifp)) {
+			d6 = D6_CSTATE(ifp);
+			if (fd != NULL) {
+				if (send_interface1(fd, ifp, d6->reason) == -1)
+					retval = -1;
+			} else
+				retval++;
+		}
 #endif
+	}
 #endif
 
 	return retval;
 }
 
-int
-script_runreason(const struct interface *ifp, const char *reason)
+static int
+script_run(struct dhcpcd_ctx *ctx, char **argv)
 {
-	struct dhcpcd_ctx *ctx = ifp->ctx;
-	char *argv[2];
 	pid_t pid;
 	int status = 0;
-	struct fd_list *fd;
-
-	if (ifp->options->script == NULL &&
-	    TAILQ_FIRST(&ifp->ctx->control_fds) == NULL)
-		return 0;
-
-	/* Make our env */
-	if (make_env(ifp, reason) == -1) {
-		logerr(__func__);
-		return -1;
-	}
-
-	if (ifp->options->script == NULL)
-		goto send_listeners;
-
-	argv[0] = ifp->options->script;
-	argv[1] = NULL;
-	logdebugx("%s: executing `%s' %s", ifp->name, argv[0], reason);
 
 	pid = script_exec(ctx, argv, ctx->script_env);
 	if (pid == -1)
@@ -652,6 +678,45 @@ script_runreason(const struct interface *ifp, const char *reason)
 			    __func__, argv[0], strsignal(WTERMSIG(status)));
 	}
 
+	return WEXITSTATUS(status);
+}
+
+int
+script_runreason(const struct interface *ifp, const char *reason)
+{
+	struct dhcpcd_ctx *ctx = ifp->ctx;
+	char *argv[2];
+	int status = 0;
+	struct fd_list *fd;
+
+	if (ifp->options->script == NULL &&
+	    TAILQ_FIRST(&ifp->ctx->control_fds) == NULL)
+		return 0;
+
+	/* Make our env */
+	if (make_env(ifp->ctx, ifp, reason) == -1) {
+		logerr(__func__);
+		return -1;
+	}
+
+	if (ifp->options->script == NULL)
+		goto send_listeners;
+
+	argv[0] = ifp->options->script;
+	argv[1] = NULL;
+	logdebugx("%s: executing `%s' %s", ifp->name, argv[0], reason);
+
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP) {
+		if (ps_root_script(ifp,
+		    ctx->script_buf, ctx->script_buflen) == -1)
+			logerr(__func__);
+		goto send_listeners;
+	}
+#endif
+
+	status = script_run(ctx, argv);
+
 send_listeners:
 	/* Send to our listeners */
 	status = 0;
@@ -665,5 +730,25 @@ send_listeners:
 			status = 1;
 	}
 
-	return WEXITSTATUS(status);
+	return status;
 }
+
+#ifdef PRIVSEP
+int
+script_runchroot(struct dhcpcd_ctx *ctx, char *script)
+{
+	char *argv[2];
+
+	/* Make our env */
+	if (make_env(ctx, NULL, "CHROOT") == -1) {
+		logerr(__func__);
+		return -1;
+	}
+
+	argv[0] = script;
+	argv[1] = NULL;
+	logdebugx("executing `%s' %s", argv[0], "CHROOT");
+
+	return script_run(ctx, argv);
+}
+#endif
